from django.core.cache import cache
from django.core.signals import request_finished
from django.test import TransactionTestCase

from modeldict import ModelDict
from modeldict.tests.models import ModelDictModel

class ModelDictTest(TransactionTestCase):
    # XXX: uses transaction test due to request_finished signal causing a rollback
    urls = 'modeldict.tests.urls'
    
    def test_api(self):
        base_count = ModelDictModel.objects.count()
        
        mydict = ModelDict(ModelDictModel, key='key', value='value')
        mydict['foo'] = 'bar'
        self.assertEquals(mydict['foo'], 'bar')
        self.assertEquals(ModelDictModel.objects.values_list('value', flat=True).get(key='foo'), 'bar')
        self.assertEquals(ModelDictModel.objects.count(), base_count+1)
        mydict['foo'] = 'bar2'
        self.assertEquals(mydict['foo'], 'bar2')
        self.assertEquals(ModelDictModel.objects.values_list('value', flat=True).get(key='foo'), 'bar2')
        self.assertEquals(ModelDictModel.objects.count(), base_count+1)
        mydict['foo2'] = 'bar'
        self.assertEquals(mydict['foo2'], 'bar')
        self.assertEquals(ModelDictModel.objects.values_list('value', flat=True).get(key='foo2'), 'bar')
        self.assertEquals(ModelDictModel.objects.count(), base_count+2)
        del mydict['foo2']
        self.assertRaises(KeyError, mydict.__getitem__, 'foo2')
        self.assertFalse(ModelDictModel.objects.filter(key='foo2').exists())
        self.assertEquals(ModelDictModel.objects.count(), base_count+1)
        
        ModelDictModel.objects.create(key='foo3', value='hello')

        self.assertEquals(mydict['foo3'], 'hello')
        self.assertTrue(ModelDictModel.objects.filter(key='foo3').exists(), True)
        self.assertEquals(ModelDictModel.objects.count(), base_count+2)
        
        request_finished.send(sender=self)
        
        self.assertEquals(mydict._cache, None)
        
        # These should still error because even though the cache repopulates (local cache)
        # the remote cache pool does not
        self.assertRaises(KeyError, mydict.__getitem__, 'foo3')
        self.assertTrue(ModelDictModel.objects.filter(key='foo3').exists())
        self.assertEquals(ModelDictModel.objects.count(), base_count+2)
        
        self.assertEquals(mydict['foo'], 'bar2')
        self.assertEquals(ModelDictModel.objects.values_list('value', flat=True).get(key='foo'), 'bar2')
        self.assertEquals(ModelDictModel.objects.count(), base_count+2)
        
        self.assertEquals(mydict.pop('foo'), 'bar2')
        self.assertEquals(mydict.pop('foo', None), None)
        self.assertFalse(ModelDictModel.objects.filter(key='foo').exists())
        self.assertEquals(ModelDictModel.objects.count(), base_count+1)
        
    def test_modeldict_instances(self):
        base_count = ModelDictModel.objects.count()
        
        mydict = ModelDict(ModelDictModel, key='key', value='value', instances=True)
        mydict['foo'] = ModelDictModel(key='foo', value='bar')
        self.assertTrue(isinstance(mydict['foo'], ModelDictModel))
        self.assertTrue(mydict['foo'].pk)
        self.assertEquals(mydict['foo'].value, 'bar')
        self.assertEquals(ModelDictModel.objects.values_list('value', flat=True).get(key='foo'), 'bar')
        self.assertEquals(ModelDictModel.objects.count(), base_count+1)
        old_pk = mydict['foo'].pk
        mydict['foo'] = ModelDictModel(key='foo', value='bar2')
        self.assertTrue(isinstance(mydict['foo'], ModelDictModel))
        self.assertEquals(mydict['foo'].pk, old_pk)
        self.assertEquals(mydict['foo'].value, 'bar2')
        self.assertEquals(ModelDictModel.objects.values_list('value', flat=True).get(key='foo'), 'bar2')
        self.assertEquals(ModelDictModel.objects.count(), base_count+1)

        # test deletion
        mydict['foo'].delete()
        self.assertTrue('foo' not in mydict)

    def test_modeldict_expirey(self):
        base_count = ModelDictModel.objects.count()
        
        mydict = ModelDict(ModelDictModel, key='key', value='value')

        self.assertEquals(mydict._cache, None)
        
        mydict['test_modeldict_expirey'] = 'hello'

        self.assertEquals(len(mydict._cache), base_count + 1)
        self.assertEquals(mydict['test_modeldict_expirey'], 'hello')

        self.client.get('/')
        
        self.assertEquals(mydict._cache, None)
        self.assertEquals(mydict['test_modeldict_expirey'], 'hello')
        self.assertEquals(len(mydict._cache), base_count + 1)
        
        request_finished.send(sender=self)

        self.assertEquals(mydict._cache, None)
        self.assertEquals(mydict['test_modeldict_expirey'], 'hello')
        self.assertEquals(len(mydict._cache), base_count + 1)

    # def test_modeldict_counts(self):
    #     # TODO: 
    #     mydict = ModelDict(ModelDictModel, key='key', value='value')
    #     mydict['test_1'] = 'foo'
    #     mydict['test_2'] = 'bar'
    #     del mydict
    #     request_finished.send(sender=self)
    # 
    #     mydict = ModelDict(ModelDictModel, key='key', value='value')
    #     # First and only cache.get() here.
    #     self.assertEqual(mydict['test_1'], 'foo')
    #     self.assertEqual(mydict['test_2'], 'bar')
    #     self.assertEqual(mydict['test_1'], 'foo')
    # 
    #     request_finished.send(sender=self)
    #     # Should not be another cache.get().
    #     self.assertEqual(mydict['test_1'], 'foo')
    #     self.assertEqual(mydict['test_2'], 'bar')
    #     self.assertEqual(mydict['test_1'], 'foo')
    # 
    #     self.assertEqual(cache._gets[c.get_key('ModelDict:ModelDictModel:key')], 1)
    #     self.assertEqual(cache._gets[c.get_key('ModelDict.last_updated:ModelDictModel:key')], 2)


import time
from threading import local

from django.db.models.signals import post_save, post_delete
from django.core.signals import request_finished
from django.core.cache import cache

try:
    VERSION = __import__('pkg_resources') \
        .get_distribution('django-modeldict').version
except Exception, e:
    VERSION = 'unknown'

NoValue = object()

class ModelDict(local):
    """
    Dictionary-style access to a model. Populates a cache and a local in-memory
    to avoid multiple hits to the database.
    
    Specifying ``instances=True`` will cause the cache to store instances rather
    than simple values.
    
    Functions in two different ways, depending on the constructor:
    
        # Given ``Model`` that has a column named ``foo`` where the value is "bar":
    
        mydict = ModelDict(Model, value='foo')
        mydict['test']
        >>> 'bar'
    
    If you want to use another key besides ``pk``, you may specify that in the
    constructor. However, this will be used as part of the cache key, so it's recommended
    to access it in the same way throughout your code.
    
        mydict = ModelDict(Model, key='foo', value='id')
        mydict['bar']
        >>> 'test'
    
    """
    def __init__(self, model, key='pk', value=None, instances=False):
        assert value is not None

        self._cache = None
        self._last_updated = None

        self.model = model
        self.key = key
        self.value = value
        self.instances = instances

        self.cache_key = 'ModelDict:%s:%s' % (model.__name__, key)
        self.last_updated_cache_key = 'ModelDict.last_updated:%s:%s' % (model.__name__, key)
        request_finished.connect(self._cleanup)
        post_save.connect(self._post_save, sender=model)
        post_delete.connect(self._post_delete, sender=model)
    
    # def __new__(cls, *args, **kwargs):
    #     self = super(ModelDict, cls).__new__(cls, *args, **kwargs)
    #     request_finished.connect(self._cleanup)
    #     return self
    
    def __getitem__(self, key):
        self._populate()
        return self._cache[key]
    
    def __setitem__(self, key, value):
        if isinstance(value, self.model):
            value = getattr(value, self.value)
        instance, created = self.model._default_manager.get_or_create(
            defaults={self.value: value},
            **{self.key: key}
        )
        if getattr(instance, self.value) != value:
            setattr(instance, self.value, value)
            instance.save()
        self._populate(reset=True)
    
    def __delitem__(self, key):
        self.model._default_manager.filter(**{self.key: key}).delete()
        self._populate(reset=True)
    
    def __len__(self):
        if self._cache is None:
            self._populate()
        return len(self._cache)
    
    def __contains__(self, key):
        self._populate()
        return key in self._cache

    def __iter__(self):
        self._populate()
        return iter(self._cache)
    
    def __repr__(self):
        return "<%s: %s>" % (self.__class__.__name__, self)

    def __str__(self):
        return ''
        return ', '.join("%s=%s" % (k, v) for k, v in self.iteritems())

    def iteritems(self):
        self._populate()
        return self._cache.iteritems()
    
    def itervalues(self):
        self._populate()
        return self._cache.itervalues()
    
    def iterkeys(self):
        self._populate()
        return self._cache.iterkeys()
    
    def items(self):
        self._populate()
        return self._cache.items()
        
    def get(self, key, default=None):
        self._populate()
        return self._cache.get(key, default)
    
    def pop(self, key, default=NoValue):
        value = self.get(key, default)
        try:
            del self[key]
        except KeyError:
            pass
        return value
    
    def setdefault(self, key, value):
        if isinstance(value, self.model):
            value = getattr(value, self.value)
        instance, created = self.model._default_manager.get_or_create(
            defaults={self.value: value},
            **{self.key: key}
        )
        self._populate(reset=True)

    def _post_save(self, sender, instance, created, **kwargs):
        if self._cache is None:
            self._populate()
        if self.instances:
            self._cache[getattr(instance, self.key)] = instance
        else:
            self._cache[getattr(instance, self.key)] = getattr(instance, self.value)
        
    def _post_delete(self, sender, instance, **kwargs):
        if self._cache is None:
            self._populate()
        self._cache.pop(getattr(instance, self.key), None)

    def _populate(self, reset=False):
        if reset:
            self._cache = None
            # TODO: Race condition in updating last_updated.  Needs
            # a test + fix.
            self.last_updated = int(time.time())
            cache.set(self.last_updated_cache_key, self.last_updated)
        elif self._cache is None:
            new_last_updated = cache.get(self.last_updated_cache_key) or 0
            if new_last_updated > (self._last_updated or 0) or \
              not getattr(self, '_cache_stale', None):
                self._cache = cache.get(self.cache_key)
                self._last_updated = new_last_updated
            else:
                self._cache = self._cache_stale
                self._cache_stale = None
        if self._cache is None:
            qs = self.model._default_manager
            if self.instances:
                self._cache = dict((getattr(i, self.key), i) for i in qs.all())
            else:
                self._cache = dict(qs.values_list(self.key, self.value))
            cache.set(self.cache_key, self._cache)
        return self._cache    

    def _cleanup(self, *args, **kwargs):
        self._cache_stale = self._cache
        self._cache = None


#!/usr/bin/env python

try:
    from setuptools import setup, find_packages
    from setuptools.command.test import test
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()
    from setuptools import setup, find_packages
    from setuptools.command.test import test

setup(
    name='django-modeldict',
    version='1.1.1',
    author='DISQUS',
    author_email='opensource@disqus.com',
    url='http://github.com/disqus/django-modeldict/',
    description = 'Stores a model as a dictionary',
    packages=find_packages(),
    zip_safe=False,
    tests_require=[
        'django',
    ],
    test_suite = 'runtests.runtests',
    include_package_data=True,
    classifiers=[
        'Framework :: Django',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Operating System :: OS Independent',
        'Topic :: Software Development'
    ],
)